import matplotlib.pyplot as plt
import numpy as np
from scipy.optimize import least_squares



def get_matrix_from_params(rot_x, rot_y, rot_z, tx, ty, tz, sx, sy, sz):
    m_rotx = np.array([[1,0,0,0], [0, np.cos(rot_x), -np.sin(rot_x), 0], [0, np.sin(rot_x), np.cos(rot_x), 0], [0,0,0,1]])
    m_roty = np.array([[np.cos(rot_y), 0, np.sin(rot_y), 0], [0,1,0,0], [-np.sin(rot_y), 0, np.cos(rot_y), 0], [0,0,0,1]])
    m_rotz = np.array([[np.cos(rot_z), -np.sin(rot_z), 0, 0], [np.sin(rot_z), np.cos(rot_z), 0, 0], [0,0,1,0], [0,0,0,1]])
    scaling_and_trans = np.array([[sx, 0, 0, tx], [0, sy, 0, ty], [0, 0, 1, 0], [0,0,0,1]])
    P = m_rotx @ m_roty @ m_rotz @ scaling_and_trans
    return P

def objective_func(x, **kwargs):
    """
    OBJECTIVE FUNCTION
    :param x: the minimization proceeds with respect to its first argument.
    :param kwargs: points 2D and 3D
    :return:
    """
    P = get_matrix_from_params(*x)

    points_2d = kwargs['points2d']
    points_3d = kwargs['points3d']
    proj = (np.matmul(P, points_3d.T)).T

    diff = proj[:, :2] - points_2d
    return diff.flatten()


def least_sq(pts2d, pts3d):
    """
    Execution algorithm least square error
    :param pts2d: 2D points
    :param pts3d: 3D points
    :return: matrix 3x4
    """
    dic = {}
    dic['points2d'] = np.array(pts2d)
    dic['points3d'] = np.array([list(pt) + [1,] for pt in pts3d])

    rot_x, rot_y, rot_z, tx, ty, tz, sx, sy, sz = 0,0,0,0,0,0,0,0,0
    initial_approx = np.array((rot_x, rot_y, rot_z, tx, ty, tz, sx, sy, sz))#np.ones(12)
    ls = least_squares(objective_func, initial_approx, verbose=0, kwargs=dic, ftol=1e-15, xtol=1e-15)#, jac='3-point')#, method="lm")

    #print(ls.x)
    P = get_matrix_from_params(*ls.x)
    return P

